/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/

#include <libc.h>
#include <stdio.h>

#include "oo7.h"
#include "globals.h"
#include "GenParams.h"
#include "BenchParams.h"

//////////////////////////////////////////////////////////////////
//
// Delete - Delete NumNewCompParts randomly chosen composite parts. 
//
//////////////////////////////////////////////////////////////////


extern o_4b	debugMode;
extern o_4b	nextCompositeId;


// JGM 93/11/8 Added basic routine to remove an element from a LinkVstr

void linkVstrRemoveFirst(LinkVstr<CompositePart>& lv,
			 Link<CompositePart> &item) {
  o_u4b size = lv.size();
  o_u4b pos = 0;
  while (pos < size) {
    if (lv[pos] == item) break;
    pos++;
  }
  // (pos==size) OR (pos<size)
  // "item not found" OR "item found"
  if (pos != size) {
    for (o_u4b i = pos; i < size-1; i++) {
      lv[i] = lv[i+1];		// shift the vstr over
    }
    lv.set_size(size-1);
  }
}



//////////////////////////////////////////////////////////////////
//
// Routine to do Delete
//
//////////////////////////////////////////////////////////////////


void delete1()
{
    // now delete the desired number of new composite parts
    for (o_u4b i = 0; i < NumNewCompParts; i++) 
    {
	o_4b	compId = nextCompositeId++;

        if (debugMode) {
	    printf("In Delete, deleting composite part %d\n", compId);
	}
	LinkVstr<CompositePart> comps =
	    PClassObject<CompositePart>::Object().select(NULL, FALSE,
		    PAttribute("CompositePart::id") == compId);

	if ( !comps.is_empty() ) 
	{
// JGM 93/11/8 Added code to fix bidirectional links relationships for
// part we are about to delete.

	  Link<CompositePart> compPart = comps[0];

	  // Remove link from document object
	  Document* doc = compPart->document;
	  doc->dirty();
	  doc->part = NULL_LINK;

	  // Remove link from atomicPart
	  o_u4b numElements = (compPart->parts).size();
	  for (o_u4b j = 0;  j < numElements; j++) {
	    AtomicPart* atomicPart = compPart->parts[j];
	    atomicPart->dirty();
	    atomicPart->partOf = NULL_LINK;
	  }

	  // Remove link from set of ComponentsPriv links
	  numElements = (compPart->usedInPriv).size();
	  for (j = 0; j < numElements; j++) {
	    BaseAssembly* baseAssemb = compPart->usedInPriv[j];
	    baseAssemb->dirty();
	    linkVstrRemoveFirst(baseAssemb->ComponentsPriv, compPart);
	  }

	  // Remove link from set of ComponentsShar links
	  numElements = (compPart->usedInShar).size();
	  for (j = 0; j < numElements; j++) {
	    BaseAssembly* baseAssemb = compPart->usedInShar[j];
	    baseAssemb->dirty();
	    linkVstrRemoveFirst(baseAssemb->ComponentsShar, compPart);
	  }

	  // Delete this instance of CompositePart
	  delete (CompositePart*)compPart;
	}
	else
            fprintf(stderr, "ERROR: Unable to locate composite part %d\n", compId);
	comps.release();
    }
}
