/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/

#include <libc.h>
#include <malloc.h>
#include <stdio.h>
#include <strings.h>

#include "oo7.h"
#include "globals.h"
#include "GenParams.h"
#include "VarParams.h"

////////////////////////////////////////////////////////////////////////////
//
// Manual Methods
//
////////////////////////////////////////////////////////////////////////////

extern o_4b	debugMode;

Implement PClassObj<Manual>; 

////////////////////////////////////////////////////////////////////////////
//
// Manual Constructor
//
////////////////////////////////////////////////////////////////////////////

Manual::Manual(o_4b modId, Link<Module> myModule)
	 : title(TitleSize), text(ManualSize), id(modId), module(myModule)
{
    o_word	curChar, strChar;
    o_4b	stringLen;
    char*	str;
    char	myText[ManualTextLength];
 
    // prepare and fill in the document title
    title.set_length(TitleSize);
    sprintf((char*)title, "Manual         %08d", modId);
    if (debugMode) {
        printf("Manual::Manual(title = %s)\n", (char*)title);
    }

    sprintf(myText, ManualText, modId);
    stringLen = strlen(myText);

    curChar = strChar = 0;
    // JGM 93/10/15 Added cast to avoid cfront 3.0 warnings
    text.set_length((int)ManualSize);
    str = (char*)text;
    while (curChar < ManualSize - 1) {
	str[curChar++] = myText[strChar++];
	if (strChar >= stringLen) { strChar = 0; }
    }
    str[curChar] = '\0';
    textLen = ManualSize;
    textSize = ManualSize;
}

// destructor
Manual::~Manual()
{
}


////////////////////////////////////////////////////////////////////////////
//
// Manual searchText Method for use in traversals
//
////////////////////////////////////////////////////////////////////////////

o_4b	Manual::searchText(const char c)
{
    if (debugMode) {
        printf("                    Manual::searchText(title = %s)\n", title);
    }

    // count occurrences of the indicated letter (for busy work)
    o_4b	count = 0;
    char*	str = (char*)text;

    while (*str != '\0') {
	if (*str++ == c) { count++; }
    }

    if (debugMode) {
        printf("                    [found %d '%c's among %d characters]\n",
	        		     count, c, str - (char*)text);
    }
    return count;
}


////////////////////////////////////////////////////////////////////////////
//
// Manual replaceText Method for use in traversals
//
////////////////////////////////////////////////////////////////////////////

o_4b	Manual::replaceText(char *oldString, char *newString)
{

    if (debugMode) {
        printf("                    Manual::changeText(title = %s)\n", title);
    }

    // check to see if the text starts with the old string

    o_4b	oldTextLength = textLen;
    o_4b	oldStrLength  = strlen(oldString);
    o_4b	foundMatch;

    // if so, change it to start with the new string instead
    foundMatch = (strncmp((char* )text, oldString,(int) oldStrLength) == 0);
    if (foundMatch) 
    {
        o_4b	newStrLength = strlen(newString);
        o_4b	lengthDiff = newStrLength - oldStrLength;

	dirty();
	if (lengthDiff > 0)
	{
	    textSize = newStrLength + 1;
	    // JGM 93/10/15 Added cast to avoid cfront 3.0 warnings
	    text.set_length((int)textSize);
	}
	textLen = newStrLength;
	// JGM 93/10/15 Added (unsigned) cast to avoid cfront 3.0
	// warnings

	strncpy((char*)text, newString, (unsigned) newStrLength);
    }

    if (debugMode) 
    {
	if (foundMatch) {
            printf("                    [changed \"%s\" to \"%s\"]\n",
	                                 oldString, newString);
	} else {
            printf("                    [no match, so no change was made]\n");
	}
    }

    return foundMatch;
}

////////////////////////////////////////////////////////////////////////////
//
// Manual firstLast Method for use in traversals
//
////////////////////////////////////////////////////////////////////////////

o_4b	Manual::firstLast()
{
    char*	str = (char*)text;

    return(str[textLen-1] == str[0]);
}
