/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/

#include <libc.h>
#include <stdio.h>

#include "oo7.h"
#include "globals.h"
#include "GenParams.h"
#include "BenchParams.h"

//////////////////////////////////////////////////////////////////
//
// Query #5 - find all base assemblies B that "use" a composite
// part with a more recent build date than B's build date.
//
//
/////////////////////////////////////////////////////////////////

extern o_4b	debugMode;

//////////////////////////////////////////////////////////////////
//
// Routine to do Query #5
//
//////////////////////////////////////////////////////////////////

o_4b query5()
{
    // scan all modules, scanning all of the base assemblies in each one,
    // scanning all of the composite parts that they use, looking for base
    // assemblies that use more recently dated composite parts

    o_4b		count = 0;
    LinkVstr<Module>	modules = PClassObject<Module>::Object().select(
					NULL, FALSE, NULL_PREDICATE);

    if (modules.is_empty()) {
	fprintf(stderr,"No Module found\n");
 	exit(1);
    }

    o_4b		size = modules.size();

    for (o_word i = 0; i < size; i++) {
	Module*		module = modules[i];

	// for each module iterate over its base assemblies
	// this assumes that we maintain a collection of base asemblies
	// in each module

    	LinkVstr<Assembly>&	baseAssemblies = module->assemblies; 
	o_4b			bsize = baseAssemblies.size();

	::dom->greadobjs(baseAssemblies, NULL);

	for (o_word j = 0; j < bsize; j++) {
	    BaseAssembly*	baseAssembly =
				(BaseAssembly*)(Assembly*)baseAssemblies[j];

            if (debugMode) {
                printf("In Query5, processing base assembly %d, buildDate = %d\n",
			baseAssembly->id, baseAssembly->buildDate);
            }

    	    LinkVstr<CompositePart>&	compositeParts =
						baseAssembly->ComponentsPriv;
	    o_4b			csize = compositeParts.size();

	    for (o_word k = 0; k < csize; k++) {
		CompositePart*	compositePart = compositeParts[k];

                if (debugMode) {
                    printf("            [Checking composite part %d, buildDate = %d]\n",
		                        compositePart->id,
					compositePart->buildDate);
                }
		if (compositePart->buildDate > baseAssembly->buildDate) {
	            baseAssembly->DoNothing();
		    count++;
	        }
	    }
	}
    }
    modules.release();
    return count;
}
